/*****************************************************************
ArduinoXbee.ino
*****************************************************************/
// incluse libreria serial SW
#include <SoftwareSerial.h>
#include <NewPing.h>
#include <EEPROM.h>
#include <stdio.h>

// define parametriche

//#define _SERIAL_DEBUG

// id station
#define ID_SLAVE_DEFAULT  1U

// id network xbee
#define ID_NETWORK        0U

// ultrasonic trig pin
#define DIST_TRIG_PIN     3
// ultrasonic echo pin
#define DIST_ECHO_PIN     2

// XBee's DOUT (TX) (Arduino's Software RX)
#define XBEE_RX_PIN       4
// XBee's DIN (RX) (Arduino's Software TX)
#define XBEE_TX_PIN       5

// light pin
#define DIST_LIGHT_PIN    6

// monitor pin
#define MONITOR_PIN       13

// max distance detectable distance ultrasonic (cm)
#define MAX_DISTANCE      300

// distance to detect object presence (default) (cm)
#define DETECT_DISTANCE_DEFAULT   100

// hysteresys for detection presence object
#define HIST_CHECK_IN     2 // n. detection to notify presence
#define HIST_CHECK_OUT    5 // n. detection to notify object leave

// parameters for led blink
#define ON_BLINK_TIME     100     // on blinck
#define PERIOD_BLINK_TIME 1000    // blink period

// parameters for repetition of xbee messages
#define MAX_REPEAT        3       // number of repetitions
#define DELAY_REPEAT      1500    // distnace between repetitions

// size of queue messages
#define MAX_CMD_QUEUE     5

// states for transisitons detections
#define TRANSIT_NONE      0
#define TRANSIT_DETECT    1
#define TRANSIT_ONGOING   2
#define TRANSIT_FINISH    3

// status for command on led
#define CMD_OFF           0
#define CMD_WAIT_ON       1
#define CMD_ON            2

// messages xbee
#define MESS_OK         "OK"
#define MESS_ON         "ON"

// system definitions for boolean
#define TRUE              1
#define FALSE             0

// range for slave id configuration
#define MIN_IDSLAVE       1
#define MAX_IDSLAVE       20

// range for distance detections
#define MIN_DISTANCEDECT  1
#define MAX_DISTANCEDECT  MAX_DISTANCE

// size of buffer data
#define BUFFER_SIZE       40


// funcions prototype
int sprintf(const char * buf,const char * format, ...);

void ReadEEPROMData();
void WriteEEPROMData();

void AddXbeeMessage(char *data);
void ParseMess(char *data);
void ParseReadXbee();

float CheckDistanceSensor();
int CheckTransitions();
int ObjectPresent();

void AddLEDCommand(int Delay, int Duration);
void SetLighState(int NewState);
int CheckCommand();
void InitXbee();

// strutture
typedef struct _comstr
{
  int PendingCmd ;
  int DelayCmd;
  int DurationCmd;
  unsigned long CmdTimer;
} CmdEle;

// variables

SoftwareSerial XBee(XBEE_RX_PIN, XBEE_TX_PIN); // RX, TX
NewPing sonar(DIST_TRIG_PIN, DIST_ECHO_PIN, MAX_DISTANCE); // NewPing setup of pins and maximum distance.

int Status = TRANSIT_NONE;
int Counter_pass = 0;
int Hist_counter = 0;
char bufferserial[BUFFER_SIZE];
char bufferserialsend[BUFFER_SIZE];
char bufferserialrcv[BUFFER_SIZE];

int PendingMessage = FALSE;
unsigned long DelayRepeat = 0;
int CountRepeat = 0;

int IdSlave = ID_SLAVE_DEFAULT;
int Distance_dect = DETECT_DISTANCE_DEFAULT;

CmdEle CommandsList[MAX_CMD_QUEUE];

char readserial[BUFFER_SIZE];
int indexserialread = 0;

int pendingxbeedata = FALSE;
int index = 0;

unsigned long BlinkTimer = 0;
int prevStateBlink = FALSE;

int log_on = FALSE;
int log_distance = FALSE;

// init parameters
void setup()
{
  // read EEPROM data
  ReadEEPROMData();
  
  // Set Xbee. This value is most important
  // for the XBee. Make sure the baud rate matches the config
  // setting of your XBee.
  XBee.begin(9600);

  // set serial log baud rate
  Serial.begin(9600);

  // set pin mode

  // pin for led activation
  pinMode(DIST_LIGHT_PIN, OUTPUT);           // set pin to input
  digitalWrite(DIST_LIGHT_PIN, LOW);         // shutdown pin

  // pin of monitor
  pinMode(MONITOR_PIN, OUTPUT);           // set pin to input
  digitalWrite(MONITOR_PIN, LOW);         // shutdown pin

  // reset command list
  memset(CommandsList,0,sizeof(CommandsList));

  // init Xbee parameters
  InitXbee();

  // init time for led blinck
  BlinkTimer = millis();
}

// read EEPROM data
void ReadEEPROMData()
{
  byte value = 0;
  int valueint = 0;
  byte valuearray[2];

  // read eeprom byte for id slave
  value = EEPROM.read(0);

  // verify if in range
  if (value >= MIN_IDSLAVE && value <= MAX_IDSLAVE)
  {
    IdSlave = value;
  }

  // read eeprom value for detection distance
  valuearray[0] = EEPROM.read(1);
  valuearray[1] = EEPROM.read(2);

  valueint = *((int*)valuearray);

  // verify if in range
  if (valueint >= MIN_DISTANCEDECT && valueint <= MAX_DISTANCEDECT)
  {
    Distance_dect = valueint;
  }  
}

// write all EEPROM data
void WriteEEPROMData()
{
  // write id slave
  EEPROM.write(0,IdSlave);

  // write distance detection
  EEPROM.write(1,((Distance_dect >> 0) & 0xFF));
  EEPROM.write(2,((Distance_dect >> 8) & 0xFF));  
}

// wait and check xbee response
int XbeeConfigResponse(char *data)
{
  unsigned long acttime = millis();
  int readval = 0;
  int index = 0;

  if (log_on)
    Serial.write("<-- ");

  // wait for a time out of 1000ms
  while(millis() - acttime < 1000)
  {
    // if data available
    if (XBee.available())
    { 
      // read the data
      readval = XBee.read();

      
      if (log_on)
        Serial.write(readval);

      // check if end message
      if (readval == '\r')
      {
        if (log_on)
          Serial.write("\n");

        // terminate the string
        data[index] = 0;

        // return ok
        return TRUE;
      }
      else
      {
        // store new byte
        data[index] = readval;

        // increase the index
        index++;
      }
    }
  }

  if (log_on)
    Serial.write("  KO\n");

  return FALSE;
}

// read and xbee parameter
int XbeeReadPar(const char *parId,unsigned int *parValue)
{
  char data[20];
  sprintf(bufferserialsend,parId);
  XBee.write(bufferserialsend);
  XBee.write(13);

  if (log_on)
  {
    Serial.write("-->");
    Serial.write(bufferserialsend);
    Serial.write("\n");
  }

  // read xbee response
  if (XbeeConfigResponse(data) == 1)
  {
    // translate parameters in a int value (note: the value are in HEX notation)
    *parValue = strtoul(data, NULL, 16);

    // return response ok
    return 1;
  }

  // reset value
  *parValue = 0;
  
  return 0;
}

// write xbee parameter
int XbeeWritePar(const char *parId,unsigned int parValue)
{
  char data[20];
  sprintf(bufferserialsend,"%s%X",parId,parValue);
  XBee.write(bufferserialsend);
  XBee.write(13);
  
  if (log_on)
  {
    Serial.write("--> ");
    Serial.write(bufferserialsend);
    Serial.write("\n");
  }

  // return Xbee response
  return XbeeConfigResponse(data);
}

// check a xbee parameter and update it if necessary
int XbeeCheckUpdatePar(const char *parId,int parValue,unsigned int *Updated)
{
  unsigned int Readval = 0;
  
  if (log_on)
  {
    Serial.write("Check: ");
    Serial.write(parId);
    Serial.write("\n");
  }

  // read actual parameter
  if (XbeeReadPar(parId,&Readval))
  {
    // if the parameter don't match
    if (Readval != parValue)
    {
      if (log_on)
      {
        Serial.write("Data read: ");
        Serial.write(Readval);
        Serial.write(" don't match\n");
      }

      // write parameter with correct value
      if (XbeeWritePar(parId,parValue) == 1)
      {
        if (log_on)
        {
          Serial.write("Data ");
          Serial.write(parId);
          Serial.write(" updated\n");
        }

        *Updated = 1;
      }
    }
    else
    {
      if (log_on)
      {
        Serial.write("Par ");
        Serial.write(parId);
        Serial.write(" match\n");
      }
    }
  
    return 1;
  }
  else
  {
    return 0;
  }
}

// set xbee in configuration mode
void XbeeeInitConfigMode()
{
  char data[20];
  if (log_on)
    Serial.write("\nEnter Xbee config mode.. \n");

  // send message for configuration mode
  XBee.write("+++");

  if (XbeeConfigResponse(data) == TRUE)
  {
    if (log_on)
      Serial.write("Xbee config mode.. OK \n");

      
    // wait for a while
    delay(1500);
      
  }
  else
  {
    if (log_on)
      Serial.write("Xbee config mode.. KO \n"); 
  }
}

// send command to store xbee parameters
void XbeeeStoreConfig()
{
  char data[20];
  if (log_on)
    Serial.write("Store Xbee config.. \n");

  // command to force store parameters
  XBee.write("ATWR");
  XBee.write(13);

  // wait answer
  if (XbeeConfigResponse(data) == TRUE)
  {
    if (log_on)
      Serial.write("Store Xbee config.. OK \n");       
  }
  else
  {
    if (log_on)
      Serial.write("Store Xbee config.. KO \n"); 
  }
}


// set xbee out of configuration mode
void XbeeeEndConfig()
{
  char data[20];
  if (log_on)
    Serial.write("End Xbee config.. \n");

  // command to exit for mconfiguration mode
  XBee.write("ATCN");
  XBee.write(13);

  // wait answer
  if (XbeeConfigResponse(data) == TRUE)
  {
    if (log_on)
      Serial.write("End Xbee config.. OK \n");       
  }
  else
  {
    if (log_on)
      Serial.write("End Xbee config.. KO \n");
  }
}


// init xbee parameters
void InitXbee()
{
  unsigned int updated = 0;

  // set monitor pin
  digitalWrite(MONITOR_PIN, HIGH);         // on pin

  // enter in configuration mode
  XbeeeInitConfigMode();

  // force id network
  XbeeCheckUpdatePar("ATID",ID_NETWORK,&updated);

  // force id slave
  XbeeCheckUpdatePar("ATNI",IdSlave,&updated);

  // force high and low destination (in broadcast mode)
  XbeeCheckUpdatePar("ATDL",0xFFFFU,&updated);
  XbeeCheckUpdatePar("ATDH",0x0000U,&updated);

  // if some parameters updated
  if (updated == 1)
  {
    // store the new configuration
    XbeeeStoreConfig();
  }

  // exit configuration mode
  XbeeeEndConfig();

  // exit from configuration mode
  digitalWrite(MONITOR_PIN, LOW);         // shutdown pin
}


// read log serial commands
void ReadSerial()
{
  int byteread;

  // if data available
  while (Serial.available())
  {
    // read the data
    byteread = Serial.read();

    // store the data
    readserial[indexserialread] = byteread;

    // enter byte: end command
    if (byteread == 13)
    {
      // terminate the string
      readserial[indexserialread] = 0;
      indexserialread = 0;

      // if command enable log
      if (!strcmp(readserial,"LOG ON"))
      {
        // set log enabled
        log_on = TRUE;
        strcpy(readserial,"LOG ENABLED\n");
      }
      // if command log off
      else if (!strcmp(readserial,"LOG OFF"))
      {
        log_on = FALSE;
        strcpy(readserial,"LOG DISABLED\n");
      }
      // if command to star led test
      else if (!strcmp(readserial,"LED TEST"))
      {
        // sen a command to on led for 4 seconds
        AddLEDCommand(0,4000);
        strcpy(readserial,"START LED TEST\n");
      }
      // enable log distance
      else if (!strcmp(readserial,"DISTANCE ON"))
      {
        log_distance = TRUE;
        strcpy(readserial,"DISTANCE LOG ENABLED\n");
      }
      // disable log distance
      else if (!strcmp(readserial,"DISTANCE OFF"))
      {
        log_distance = FALSE;
        strcpy(readserial,"DISTANCE LOG DISABLED\n");
      }
      // read actual id
      else if (!strcmp(readserial,"GET ID"))
      {
        //return actual stored id
        sprintf(readserial,"ID: %d\n",IdSlave);
      }
      // read actual stored distance
      else if (!strcmp(readserial,"GET DISTANCE"))
      {
        // return the sotred value
        sprintf(readserial,"DISTANCE: %d CM\n",Distance_dect);
      }
      // set station id
      else if (!strncmp(readserial,"SET ID ",7))
      {
        // read new vslue
        int id = atoi(&(readserial[7]));

        // check if value in the range
        if (id >= MIN_IDSLAVE && id <= MAX_IDSLAVE)
        {
          // store the new value internally and in EEPROM
          IdSlave = id;
          WriteEEPROMData();
          
          sprintf(readserial,"NEW ID: %d SET (REBOOT TO APPLY)\n",id);
        }
        else
        {
          strcpy(readserial,"ID NOT VALID (RANGE 1-20)\n");
        }
      }
      // set distance detection
      else if (!strncmp(readserial,"SET DISTANCE ",13))
      {
        // get the new value
        int distance = atoi(&(readserial[13]));

        // check if in range
        if (distance >= MIN_DISTANCEDECT && distance <= MAX_DISTANCEDECT)
        {
          // store internally adn in EEPROM
          Distance_dect = distance;
          WriteEEPROMData();
          
          sprintf(readserial,"NEW DISTANCE: %d CM SET\n",distance);
        }
        else
        {
          strcpy(readserial,"DISTANCE NOT VALID (RANGE 1-400)\n");
        }
      }
      // unknow command
      else
      {
        strcpy(readserial,"UNKNOW COMMAND\n");
      }

      // returne the answer
      Serial.print(readserial);
    }
    else
    {
      // store the byte in buffer
      readserial[indexserialread] = byteread;
      indexserialread = (indexserialread + 1) % BUFFER_SIZE;
    }   
  }
}


// min loop
void loop()
{
  // check transition of objects
  CheckTransitions();

  // send xbbe messages
  SendXbeeMessage();

  // parse read xbee commands
  ParseReadXbee();

  // check pending commands (i.e led on)
  CheckCommand();

  // read serial commands
  ReadSerial();

  // delay (for ultrasonic sensor stabilitation)
  delay(40);                     // Wait 50ms between pings (about 20 pings/sec). 29ms should be the shortest delay between pings.
}

// read and parse xbee message
void ParseReadXbee()
{
  int readval = 0;

  // if new data available
  while (XBee.available())
  { 
    // read the data
    readval = XBee.read();

    // if init message
    if (readval == 0x02)
    {
      // start message store
      index = 0;
      pendingxbeedata = TRUE;
    }
    else
    {
      // command pending
      if (pendingxbeedata == TRUE)
      {
        // read if end message
        if (readval == 0x03)
        {
          // terminate the string
          bufferserialrcv[index] = 0;

          // if some data read
          if (index > 0)
          {
            if (log_on)
            {
              Serial.write("Message: ");
              Serial.write(bufferserialrcv);
              Serial.write("\n");
            }

            // parse the message
            ParseMess(bufferserialrcv);
          }

          // clear pending message
          pendingxbeedata = FALSE;
        }
        else  // not end message
        { 
          // if within max message sixe
          if (index < BUFFER_SIZE - 1)
          {
            // sotre the data
            bufferserialrcv[index] = readval;
            index++;
          }
          else
          {
            // clear pending message
            pendingxbeedata = FALSE;
            break;
          }
        }
      }
    } 
  }
}

// add a ne led command in queue
void AddLEDCommand(int Delay, int Duration)
{
  int i;

  // scna all the queue
  for ( i = 0; i < MAX_CMD_QUEUE ; i++)
  {
    // if a empty slot in queue
    if (CommandsList[i].PendingCmd == CMD_OFF)
    {
      // store command parameters
      CommandsList[i].DelayCmd = Delay;
      CommandsList[i].DurationCmd = Duration;
      CommandsList[i].CmdTimer = millis();
      CommandsList[i].PendingCmd = CMD_WAIT_ON;
  
      break;
    }
  }
}

// parse xbee message
void ParseMess(char *data)
{
  char *act = data;
  char *endidsource = strchr (act,' ');
  char *initidsource = act;

  // the message start with source id
  if (endidsource != NULL)
  {
    // read the id
    endidsource[0] = 0;
    act = endidsource + 1;
    int IdSender = atoi(initidsource);
    char *initiddest = act;

    if (log_on)
    {
      Serial.write(initidsource); 
      Serial.write(" source\n");
    }  

    // if sende is a master
    if (IdSender <= 0)
    {      
      char *endiddest = strchr(act,' ');

      // check if destiation id
      if (endiddest != NULL)
      {
        endiddest[0] = 0;
        act = endiddest + 1; 
        if (log_on)
        {
          Serial.write(initiddest); 
          Serial.write(" dest\n");  
        }

        int IdDest = atoi(initiddest);
        char *initimess = act;

        // if destitation is this slave
        if (IdDest == IdSlave)
        {
          char *endmess = strchr(act,' ');
     
          if (endmess != NULL)
          {
            endmess[0]= 0;
            act = endmess + 1;

            if (log_on)
            {
              Serial.write(initimess); 
              Serial.write(" cmd\n");
            }

            // command: handshake
            if (strcmp(initimess,MESS_OK) == 0)
            {
              if (log_on)
                Serial.write("DONE\n");

              // reset flag pending trasmission
              PendingMessage = FALSE;
            }
            else
            {
              // if command led on
              if (strcmp(initimess,MESS_ON) == 0)
              {
                char *initidelay = act;
                char *enddelay = strchr(act,' ');

                //read command parameters
                if (enddelay != NULL)
                {
                  enddelay[0] = 0;
                  act = enddelay + 1;
                  int Delay = atoi(initidelay);
                  char *initduration = act;

                  if (log_on)
                  {
                    Serial.write(initidelay); 
                    Serial.write(" delay\n");
                  }
                
                  char *endduration = strchr(act,' ');
                  
                  if (endduration != NULL)
                  {
                    endduration[0] = 0;
                    int Duration = atoi(initduration);

                    if (log_on)
                    {
                      Serial.write(initduration); 
                      Serial.write(" duration\n");
                    }

                    // store the command in queue
                    AddLEDCommand(Delay,Duration);
                  }
                }
              }
            }
          }
        }
      }
    }
  }
}

// send messages to xbee
void SendXbeeMessage()
{
  unsigned long acttime = millis();

  // if a message pending
  if (PendingMessage == TRUE)
  {
    // if first send or elapsed timeout for repetition
    if (CountRepeat == 0 || acttime - DelayRepeat > DELAY_REPEAT)
    {
      if (log_on)
      {
        Serial.write(bufferserialsend);
        Serial.write("\n");
      }

      // send the command
      XBee.write(bufferserialsend);

      // store the time to send command
      DelayRepeat = acttime;

      // if elapsed number of repetition
      if (++CountRepeat >= MAX_REPEAT)
      {
        // clear the message
        PendingMessage = FALSE;
      }
    }
  }   
}

// add a new command to send to xbee
void AddXbeeMessage(char *data)
{
  // copy the data
  strcpy(bufferserialsend,data);

  // init parameters of message
  DelayRepeat = millis();
  CountRepeat = 0;
  PendingMessage = TRUE;
}

// support fuction to compose string
int sprintf(char * buf,const char * format, ...)
{
  int len;
  va_list arglist;

  va_start(arglist, format);

  len = vsprintf(buf, format, arglist);

  va_end(arglist);

  return len;
}

// detect if a object present
int ObjectPresent()
{
  int detect = FALSE;
  
   // riead actual distance
  int cm = sonar.ping_cm();

  // if a object in the range
  if (cm > 0 && cm < Distance_dect)
  {
    // set object detected
    detect = TRUE;
  }

  if (log_distance == TRUE)
  {
    Serial.print("Distance: ");
    if (cm > 0)
    {
      Serial.print(cm); // Send ping, get distance in cm and print result (0 = outside set distance range)
      Serial.print("cm");
      Serial.print(" DETECT: ");
      if (detect)
      {
        Serial.println("YES");
      }
      else
      {
        Serial.println("NO");
      }
    }
    else
    {
      Serial.println(">max DETECT: no"); // Send ping, get distance in cm and print result (0 = outside set distance range)
    }
    
  }

  // return object presence
  return detect;
}

// set the state of led
void SetLighState(int NewState)
{
  unsigned long acttime = millis();
  int Elapsed = acttime - BlinkTimer;

  // check if on phase
  if (Elapsed < ON_BLINK_TIME)
  {
    // if not already on
    if (prevStateBlink == TRUE)
    {
      // set pin on
      digitalWrite(DIST_LIGHT_PIN, HIGH);
    }
  }
  else
  {
    // if elapsed period
    if (Elapsed > PERIOD_BLINK_TIME)
    {
      // store init of new period
      BlinkTimer = acttime;

      // store the new state
      prevStateBlink = NewState;
    }

    // set pin off
    digitalWrite(DIST_LIGHT_PIN, LOW);
  }
}

// manage commands
int CheckCommand()
{
  int i;
  int LedState = 0;
  unsigned long acttime = millis();

  // check all queue
  for ( i = 0; i < MAX_CMD_QUEUE ; i++)
  {
    // check the state of element
    switch (CommandsList[i].PendingCmd)
    {
      // state: off, notinhg to do
      case CMD_OFF: 
        break;
        
      // state: wiat on
      case CMD_WAIT_ON: 

        // if elapsed delay time
        if (acttime - CommandsList[i].CmdTimer >= CommandsList[i].DelayCmd)
        {
          // store the act time
          CommandsList[i].CmdTimer = acttime;

          // set command
          CommandsList[i].PendingCmd = CMD_ON;

          if (log_on)
            Serial.write(" LIGHT: ON\n");

        }
  
        break;
        
      // state: on
      case CMD_ON: 

        // set state on
        LedState = 1;

        // if elapsed time for command
        if (acttime - CommandsList[i].CmdTimer >= CommandsList[i].DurationCmd)
        {
          // reset command
          CommandsList[i].PendingCmd = CMD_OFF;

          if (log_on)
            Serial.write(" LIGHT: OFF\n");

        }
        break;

      // invalid: set defualt
      default:
        CommandsList[i].PendingCmd = CMD_OFF;
        break;
    }
  }

  // update the led state with new command
  SetLighState(LedState);
}

// check transition to detect objectj
int CheckTransitions()
{
  // check the status
  switch (Status)
  {
    // no object
    case TRANSIT_NONE: 

      // if object present
      if (ObjectPresent())
      {
        // reset the counter
        Hist_counter = 0;

        //  set next state transition pending
        Status = TRANSIT_DETECT;
      }

      // set monitor pin off
      digitalWrite(MONITOR_PIN, LOW);

      break;

    case TRANSIT_DETECT: 

      // if obect present
      if (ObjectPresent())
      {
        // increase the counter
        Hist_counter++;
      }
      else
      {
        // decrease the counter
        Hist_counter--;
      }

      // if elapsed hyteresis
      if (Hist_counter >= HIST_CHECK_IN)
      {
        // reset the counter
        Hist_counter = 0;

        // set status transition ongoing
        Status = TRANSIT_ONGOING;
      }
      else
      {
        // if counter 0
        if (Hist_counter <= 0)
        {
          // come back to no transition
          Status = TRANSIT_NONE;
        }
      }
      break;

    // transition ongoing
    case TRANSIT_ONGOING: 

      // check if obect present
      if (ObjectPresent())
      {
        // reset counter
        Hist_counter = 0;
      }
      else
      {
        // increase the counter
        Hist_counter++;
      }

      // if enough time no obect detect
      if (Hist_counter >= HIST_CHECK_OUT)
      {   
        // increase the statistic
        Counter_pass++;

        // set status transition end
        Status = TRANSIT_FINISH;
        
      }

      // reset monitor pin
      digitalWrite(MONITOR_PIN, HIGH); 
      
      break;

    // transition elapsed
    case TRANSIT_FINISH:

      //reset counter
      Hist_counter = 0;
      
      // set status no transition
      Status = TRANSIT_NONE;

      // create new message for xbee
      sprintf(bufferserial,"\002%d PASS: %d\003",IdSlave,Counter_pass);

      // send notification to xbee: new transition
      AddXbeeMessage(bufferserial);
      
      break;

    // not valid state: reset all
    default:

      Status = TRANSIT_NONE;
      Hist_counter = 0;
      break;
  }

  // retur actual state
  return Status;
}
